'''
Little helper script to execute the program stored on a Galil controller

This is useful to reinitialize the mesoSPIM Galil handcontroller after 
the program execution stopped.
'''
self.execute_galil_program()